import {LocationStrategy} from './location_strategy';
import {StringWrapper, isPresent, CONST_EXPR} from 'angular2/src/core/facade/lang';
import {EventEmitter, ObservableWrapper} from 'angular2/src/core/facade/async';
import {isBlank} from 'angular2/src/core/facade/lang';
import {BaseException, WrappedException} from 'angular2/src/core/facade/exceptions';
import {OpaqueToken, Injectable, Optional, Inject} from 'angular2/src/core/di';

/**
 * The `APP_BASE_HREF` token represents the base href to be used with the
 * {@link PathLocationStrategy}.
 *
 * If you're using {@link PathLocationStrategy}, you must provide a binding to a string
 * representing the URL prefix that should be preserved when generating and recognizing
 * URLs.
 *
 * ## Example
 *
 * ```
 * import {Component, View} from 'angular2/angular2';
 * import {ROUTER_DIRECTIVES, routerBindings, RouteConfig} from 'angular2/router';
 *
 * @Component({...})
 * @View({directives: [ROUTER_DIRECTIVES]})
 * @RouteConfig([
 *  {...},
 * ])
 * class AppCmp {
 *   // ...
 * }
 *
 * bootstrap(AppCmp, [
 *   routerBindings(AppCmp),
 *   PathLocationStrategy,
 *   bind(APP_BASE_HREF).toValue('/my/app')
 * ]);
 * ```
 */
export const APP_BASE_HREF: OpaqueToken = CONST_EXPR(new OpaqueToken('appBaseHref'));

/**
 * `Location` is a service that applications can use to interact with a browser's URL.
 * Depending on which {@link LocationStrategy} is used, `Location` will either persist
 * to the URL's path or the URL's hash segment.
 *
 * Note: it's better to use {@link Router#navigate} service to trigger route changes. Use
 * `Location` only if you need to interact with or create normalized URLs outside of
 * routing.
 *
 * `Location` is responsible for normalizing the URL against the application's base href.
 * A normalized URL is absolute from the URL host, includes the application's base href, and has no
 * trailing slash:
 * - `/my/app/user/123` is normalized
 * - `my/app/user/123` **is not** normalized
 * - `/my/app/user/123/` **is not** normalized
 *
 * ## Example
 *
 * ```
 * import {Component, View} from 'angular2/angular2';
 * import {
 *   ROUTER_DIRECTIVES,
 *   routerBindings,
 *   RouteConfig,
 *   Location
 * } from 'angular2/router';
 *
 * @Component({...})
 * @View({directives: [ROUTER_DIRECTIVES]})
 * @RouteConfig([
 *  {...},
 * ])
 * class AppCmp {
 *   constructor(location: Location) {
 *     location.go('/foo');
 *   }
 * }
 *
 * bootstrap(AppCmp, [routerBindings(AppCmp)]);
 * ```
 */
@Injectable()
export class Location {
  _subject: EventEmitter = new EventEmitter();
  _baseHref: string;

  constructor(public platformStrategy: LocationStrategy,
              @Optional() @Inject(APP_BASE_HREF) href?: string) {
    var browserBaseHref = isPresent(href) ? href : this.platformStrategy.getBaseHref();

    if (isBlank(browserBaseHref)) {
      throw new BaseException(
          `No base href set. Either provide a binding to "appBaseHrefToken" or add a base element.`);
    }

    this._baseHref = stripTrailingSlash(stripIndexHtml(browserBaseHref));
    this.platformStrategy.onPopState(
        (_) => { ObservableWrapper.callNext(this._subject, {'url': this.path(), 'pop': true}); });
  }

  /**
   * Returns the normalized URL path.
   */
  path(): string { return this.normalize(this.platformStrategy.path()); }

  /**
   * Given a string representing a URL, returns the normalized URL path.
   */
  normalize(url: string): string {
    return stripTrailingSlash(_stripBaseHref(this._baseHref, stripIndexHtml(url)));
  }

  /**
   * Given a string representing a URL, returns the normalized URL path.
   * If the given URL doesn't begin with a leading slash (`'/'`), this method adds one
   * before normalizing.
   */
  normalizeAbsolutely(url: string): string {
    if (!url.startsWith('/')) {
      url = '/' + url;
    }
    return stripTrailingSlash(_addBaseHref(this._baseHref, url));
  }

  /**
   * Changes the browsers URL to the normalized version of the given URL, and pushes a
   * new item onto the platform's history.
   */
  go(url: string): void {
    var finalUrl = this.normalizeAbsolutely(url);
    this.platformStrategy.pushState(null, '', finalUrl);
  }

  /**
   * Navigates forward in the platform's history.
   */
  forward(): void { this.platformStrategy.forward(); }

  /**
   * Navigates back in the platform's history.
   */
  back(): void { this.platformStrategy.back(); }

  /**
   * Subscribe to the platform's `popState` events.
   */
  subscribe(onNext: (value: any) => void, onThrow: (exception: any) => void = null,
            onReturn: () => void = null): void {
    ObservableWrapper.subscribe(this._subject, onNext, onThrow, onReturn);
  }
}

function _stripBaseHref(baseHref: string, url: string): string {
  if (baseHref.length > 0 && url.startsWith(baseHref)) {
    return url.substring(baseHref.length);
  }
  return url;
}

function _addBaseHref(baseHref: string, url: string): string {
  if (!url.startsWith(baseHref)) {
    return baseHref + url;
  }
  return url;
}

function stripIndexHtml(url: string): string {
  if (/\/index.html$/g.test(url)) {
    // '/index.html'.length == 11
    return url.substring(0, url.length - 11);
  }
  return url;
}

function stripTrailingSlash(url: string): string {
  if (/\/$/g.test(url)) {
    url = url.substring(0, url.length - 1);
  }
  return url;
}
